import os
import pandas as pd
from sklearn.model_selection import train_test_split
from datasets import Dataset, load_from_disk
from transformers import BertTokenizer, BertPreTrainedModel, BertModel
from torch import nn
import torch
from peft import get_peft_model, LoraConfig, TaskType
from transformers import Trainer, TrainingArguments
import numpy as np
from sklearn.metrics import f1_score, precision_score, recall_score
from sklearn.utils.class_weight import compute_class_weight

torch.cuda.empty_cache()

# Определяем устройство (GPU или CPU)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# Пути для сохранения токенизированных данных
TOKENIZED_DATA_DIR = "./tokenized_data"
TRAIN_TOKENIZED_PATH = os.path.join(TOKENIZED_DATA_DIR, "train")
VAL_TOKENIZED_PATH = os.path.join(TOKENIZED_DATA_DIR, "val")
TEST_TOKENIZED_PATH = os.path.join(TOKENIZED_DATA_DIR, "test")

# Загрузка данных
data = pd.read_csv('all_dataset.csv')

# Разделение данных на train, validation и test
train_data, test_data = train_test_split(data, test_size=0.2, random_state=42)
train_data, val_data = train_test_split(train_data, test_size=0.1, random_state=42)

# Преобразование данных в формат Dataset
train_dataset = Dataset.from_pandas(train_data)
val_dataset = Dataset.from_pandas(val_data)
test_dataset = Dataset.from_pandas(test_data)

# Загрузка токенизатора
tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')

# Функция для токенизации
def preprocess_function(examples):
    tokenized = tokenizer(examples['prompt'], truncation=True, padding=True, max_length=512)
    # Преобразуем метки в числовой формат
    labels_safety = [0 if label == "safe" else 1 for label in examples['safety']]
    labels_attack = [0 if label == "jailbreak" else 1 if label == "evasion" else 2 if label == "generic attack" else 3 for label in examples['type']]
    # Объединяем метки в список
    tokenized['labels'] = list(map(lambda x, y: [x, y], labels_safety, labels_attack))
    return tokenized

# Если токенизированные данные уже сохранены, загружаем их, иначе выполняем токенизацию
if os.path.exists(TRAIN_TOKENIZED_PATH) and os.path.exists(VAL_TOKENIZED_PATH) and os.path.exists(TEST_TOKENIZED_PATH):
    print("Загрузка токенизированных данных с диска...")
    train_dataset = load_from_disk(TRAIN_TOKENIZED_PATH)
    val_dataset = load_from_disk(VAL_TOKENIZED_PATH)
    test_dataset = load_from_disk(TEST_TOKENIZED_PATH)
else:
    print("Токенизация данных...")
    train_dataset = train_dataset.map(preprocess_function, batched=True)
    val_dataset = val_dataset.map(preprocess_function, batched=True)
    test_dataset = test_dataset.map(preprocess_function, batched=True)
    os.makedirs(TOKENIZED_DATA_DIR, exist_ok=True)
    train_dataset.save_to_disk(TRAIN_TOKENIZED_PATH)
    val_dataset.save_to_disk(VAL_TOKENIZED_PATH)
    test_dataset.save_to_disk(TEST_TOKENIZED_PATH)
    print("Токенизированные данные сохранены на диск.")

# Вычисление весов классов для обеих задач
classes_task1 = np.unique(train_data['safety'])
class_weights_task1 = compute_class_weight('balanced', classes=classes_task1, y=train_data['safety'])
class_weights_dict_task1 = {i: weight for i, weight in enumerate(class_weights_task1)}

classes_task2 = np.unique(train_data[train_data['safety'] == 'unsafe']['type'])
class_weights_task2 = compute_class_weight('balanced', classes=classes_task2, y=train_data[train_data['safety'] == 'unsafe']['type'])
class_weights_dict_task2 = {i: weight for i, weight in enumerate(class_weights_task2)}

# Перенос весов на устройство
class_weights_task1_tensor = torch.tensor(list(class_weights_dict_task1.values()), dtype=torch.float32).to(device)
class_weights_task2_tensor = torch.tensor(list(class_weights_dict_task2.values()), dtype=torch.float32).to(device)

# Определение модели
class MultiTaskBert(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.bert = BertModel(config)
        self.classifier_safety = nn.Linear(config.hidden_size, 2)  # safe/unsafe
        self.classifier_attack = nn.Linear(config.hidden_size, 4)  # jailbreak, evasion, generic attack, injection

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        **kwargs
    ):
        # Перенос входных данных на устройство
        if input_ids is not None:
            input_ids = input_ids.to(device)
        if attention_mask is not None:
            attention_mask = attention_mask.to(device)
        if labels is not None:
            if not torch.is_tensor(labels):
                labels = torch.tensor(labels, dtype=torch.long).to(device)
            else:
                labels = labels.to(device)

        # Запуск BERT
        outputs = None
        if inputs_embeds is not None:
            outputs = self.bert(
                inputs_embeds=inputs_embeds,
                attention_mask=attention_mask,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=True,
            )
        else:
            outputs = self.bert(
                input_ids=input_ids,
                attention_mask=attention_mask,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=True,
            )

        pooled_output = outputs.last_hidden_state[:, 0, :]  # Используем [CLS] токен

        # Вычисляем логиты для каждой задачи
        logits_safety = self.classifier_safety(pooled_output)
        logits_attack = self.classifier_attack(pooled_output)

        # Объединяем логиты в один тензор
        logits = torch.cat([logits_safety, logits_attack], dim=1)

        loss = None
        if labels is not None:
            labels_safety = labels[:, 0]
            labels_attack = labels[:, 1]
            loss_fct_safety = nn.CrossEntropyLoss(weight=class_weights_task1_tensor)
            loss_fct_attack = nn.CrossEntropyLoss(weight=class_weights_task2_tensor)
            loss_safety = loss_fct_safety(logits_safety, labels_safety)
            loss_attack = loss_fct_attack(logits_attack, labels_attack)
            loss = loss_safety + loss_attack

        # Формируем возвращаемый словарь: возвращаем только loss и logits, исключая поля, которые могут быть None
        return {"loss": loss, "logits": logits}

# Загрузка модели
model = MultiTaskBert.from_pretrained('bert-base-uncased')
model = model.to(device)

# Конфигурация LoRA
lora_config = LoraConfig(
    task_type=TaskType.SEQ_CLS,
    r=8,
    lora_alpha=32,
    lora_dropout=0.1,
    target_modules=["query", "value"],
)

# Добавление LoRA к модели
model = get_peft_model(model, lora_config)
model.print_trainable_parameters()

# Функция для вычисления метрик
def compute_metrics(p):
    # Разбиваем объединённый массив логитов: первые 2 значения для safe/unsafe, следующие 4 для типа атаки
    safety_logits = p.predictions[:, :2]
    attack_logits = p.predictions[:, 2:]
    preds_safety = np.argmax(safety_logits, axis=1)
    preds_attack = np.argmax(attack_logits, axis=1)
    
    labels_safety = p.label_ids[:, 0]
    labels_attack = p.label_ids[:, 1]
    
    f1_safety = f1_score(labels_safety, preds_safety, average='weighted')
    precision_safety = precision_score(labels_safety, preds_safety, average='weighted')
    recall_safety = recall_score(labels_safety, preds_safety, average='weighted')
    
    f1_attack = f1_score(labels_attack, preds_attack, average='weighted')
    precision_attack = precision_score(labels_attack, preds_attack, average='weighted')
    recall_attack = recall_score(labels_attack, preds_attack, average='weighted')
    
    return {
        'f1_safety': f1_safety,
        'precision_safety': precision_safety,
        'recall_safety': recall_safety,
        'f1_attack': f1_attack,
        'precision_attack': precision_attack,
        'recall_attack': recall_attack,
    }

# Настройка параметров обучения
training_args = TrainingArguments(
    output_dir='./results',
    eval_strategy="epoch",
    save_strategy="epoch",
    learning_rate=5e-6,
    per_device_train_batch_size=16,
    per_device_eval_batch_size=16,
    num_train_epochs=3,
    weight_decay=0.01,
    logging_dir='./logs',
    logging_steps=10,
    save_total_limit=2,
    load_best_model_at_end=True,
    metric_for_best_model="f1_safety",
    greater_is_better=True,
    fp16=True,  # При необходимости можно отключить: fp16=False
    max_grad_norm=1.0,
    warmup_steps=100,
    report_to="none",
)

# Создание Trainer
trainer = Trainer(
    model=model,
    args=training_args,
    train_dataset=train_dataset,
    eval_dataset=val_dataset,
    compute_metrics=compute_metrics,
)

# Обучение модели
trainer.train()

# Оценка модели на тестовом наборе
results = trainer.evaluate(test_dataset)
print("Fine-tuned Model Evaluation Results:")
print(results)

# Сохранение модели и токенизатора
model.save_pretrained('./fine-tuned-bert-lora-multi-task')
tokenizer.save_pretrained('./fine-tuned-bert-lora-multi-task')