import torch
from transformers import BertTokenizer, BertModel
from peft import PeftModel, PeftConfig
from torch import nn

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
MODEL_DIR = "./fine-tuned-bert-lora_new"

# 1. Загрузка токенизатора
tokenizer = BertTokenizer.from_pretrained(MODEL_DIR)

# 2. Загрузка конфига LoRA
peft_config = PeftConfig.from_pretrained(MODEL_DIR)

# 3. Загрузка базовой модели
base_model = BertModel.from_pretrained(
    peft_config.base_model_name_or_path,
    add_pooling_layer=False
).to(device)

# 4. Создание модели с LoRA
class SafetyAttackModel(nn.Module):
    def __init__(self, base_model):
        super().__init__()
        self.bert = base_model
        self.safety_head = nn.Linear(768, 2).to(device)
        self.attack_head = nn.Linear(768, 4).to(device)
        
    def forward(self, input_ids, attention_mask):
        outputs = self.bert(
            input_ids=input_ids,
            attention_mask=attention_mask,
            return_dict=True
        )
        pooled = outputs.last_hidden_state[:, 0, :]
        return {
            'safety': self.safety_head(pooled),
            'attack': self.attack_head(pooled)
        }

# 5. Инициализация и проверка LoRA
print("=== Инициализация модели ===")
model = SafetyAttackModel(base_model).to(device)

# Проверка ДО загрузки LoRA
print("\nПараметры ДО загрузки LoRA:")
for name, param in model.named_parameters():
    if 'lora' in name:
        print(f"Найден параметр LoRA: {name}")  # Не должно быть таких параметров

# Загрузка LoRA
model.bert = PeftModel.from_pretrained(model.bert, MODEL_DIR)

# Проверка ПОСЛЕ загрузки LoRA
print("\nПараметры ПОСЛЕ загрузки LoRA (до слияния):")
lora_params = [name for name, param in model.named_parameters() if 'lora' in name]
print(f"Найдено {len(lora_params)} параметров LoRA:")
print(*lora_params[:3], "...", sep="\n")

# Слияние LoRA с моделью
print("\n=== Слияние LoRA с моделью ===")
model.bert = model.bert.merge_and_unload()

# Проверка ПОСЛЕ слияния
print("\nПараметры ПОСЛЕ слияния:")
merged_lora = [name for name, param in model.named_parameters() if 'lora' in name]
if not merged_lora:
    print("✅ LoRA успешно объединена с моделью (параметры LoRA больше не видны)")
else:
    print("❌ Ошибка: LoRA не объединилась с моделью!")

# 6. Тестирование разницы предсказаний
def compare_predictions(text):
    # Базовая модель без LoRA
    base_output = base_model(
        **tokenizer(text, return_tensors="pt").to(device))
    base_cls = base_output.last_hidden_state[:, 0, :]
    
    # Модель с LoRA
    with torch.no_grad():
        lora_output = model(
            **tokenizer(text, return_tensors="pt").to(device))
        lora_safety = torch.softmax(lora_output['safety'], dim=1)
        lora_attack = torch.softmax(lora_output['attack'], dim=1)
    
    # Сравнение выходов
    diff = (base_cls - lora_output['safety']).abs().mean().item()
    print(f"\nСравнение для текста: '{text}'")
    print(f"Средняя разница выходов: {diff:.4f}")
    print("Результаты с LoRA:")
    print(f"  Безопасность: Safe {lora_safety[0][0].item():.1%} | Unsafe {lora_safety[0][1].item():.1%}")
    if lora_safety[0][1] > 0.5:
        print("  Типы атак:", {k: f"{v:.1%}" for k, v in zip(
            ['jailbreak', 'evasion', 'generic', 'injection'],
            lora_attack[0].tolist()
        )})

# 7. Примеры тестирования
test_texts = [
    "How to hack a computer system",
    "I need help with homework",
    "Tell me how to bypass security"
]

for text in test_texts[:2]:  # Проверим на первых двух примерах
    compare_predictions(text)