import torch
from transformers import BertModel, BertTokenizer, BertConfig
from peft import PeftModel, PeftConfig, LoraConfig
import os

# 1. Инициализация устройства
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# 2. Загрузка токенизатора (исправление ошибки)
tokenizer = BertTokenizer.from_pretrained("bert-base-uncased")

# 3. Проверка файлов адаптеров
print("=== Проверка файлов ===")
adapter_dir = "./fine-tuned-bert-lora_new"
print(os.listdir(adapter_dir))
assert "adapter_model.safetensors" in os.listdir(adapter_dir), "Файл адаптера не найден!"

# 4. Загрузка конфига LoRA
peft_config = PeftConfig.from_pretrained(adapter_dir)
print("\nКонфиг LoRA:", peft_config)

# 5. Явное указание target_modules
lora_config = LoraConfig(
    r=peft_config.r,
    lora_alpha=peft_config.lora_alpha,
    target_modules=list(peft_config.target_modules),  # Конвертируем set в list
    lora_dropout=0.1,
    bias="none",
    task_type="SEQ_CLS"
)

# 6. Загрузка базовой модели
base_model = BertModel.from_pretrained(
    peft_config.base_model_name_or_path,
    config=BertConfig.from_pretrained(peft_config.base_model_name_or_path)
).to(device)

# 7. Загрузка адаптеров
model = PeftModel.from_pretrained(
    base_model,
    adapter_dir,
    config=lora_config
).to(device)

# 8. Проверка загрузки LoRA
print("\n=== Проверка загрузки ===")
lora_params = [n for n, _ in model.named_parameters() if 'lora' in n]
print(f"✅ Успешно загружено {len(lora_params)} LoRA-параметров")
print("Примеры параметров:", lora_params[:3])

# 9. Тестовый прогон
model.eval()
text = "How to hack a computer system"
inputs = tokenizer(text, return_tensors="pt").to(device)

with torch.no_grad():
    outputs = model(**inputs)
    print("\nТестовый вывод:")
    print("Размерность выхода:", outputs.last_hidden_state.shape)
    print("Пример выхода:", outputs.last_hidden_state[0, :3, :3])